// ===================================================
// 
//
// ===================================================

#include "arduino.h"

#include "pins.h"
#include "settings.h"
#include "avr/sleep.h"
#include <avr/pgmspace.h>
#include <avr/power.h>

#include "morse.h"
#include "memory.h"
#include "keyerio.h"

volatile bool g_DitPressed=false;
volatile bool g_DahPressed=false;
volatile int  g_ButtonPressed=0;

volatile bool g_IsInCommandMode = false;
volatile bool g_IsInProgrammingMode = false;

volatile unsigned int g_CommandSequence = 0;
volatile unsigned long int g_SymbolEndTime = 0;

volatile int g_PowerCount = POWER_DOWN_TIME;
volatile int g_SaveCount = POWER_SAVE_TIME;

Morse morse;
Memory keyerMem;
KeyerIO keyerIO;

KeyerIO* keyer;

void setup()
{
  pinMode(DIT_PIN, INPUT_PULLUP);
  pinMode(DAH_PIN, INPUT_PULLUP);
  
  pinMode(WPM_POTI_PIN, INPUT );
  pinMode(BUTTONS_PIN, INPUT );
  
  pinMode(OUT_PIN, OUTPUT);
  pinMode(BEEPER_PIN, OUTPUT);
  pinMode(COMMAND_LED_PIN, OUTPUT);
  pinMode(POWER_LED_PIN, OUTPUT);

  digitalWrite( POWER_LED_PIN, true );
  
  bool reset = false;
  if( EEPROM.read( EEPROM_WPM_ADDRESS ) == 0xff ||
      EEPROM.read( EEPROM_MEM_OFFSET  ) == 0xff ||
      keyerIO.checkButtons() == 1 )
  {
    reset = true;
  }
  
  keyerMem.init(reset);
  keyerIO.init (reset);

  delay( 100 );
  keyerIO.bleepUp();

  // wait for the user to release any button if pressed
  do
  { 
    delay(100);
  }
  while( keyerIO.checkButtons() != 0 );
  
  attachInterrupt( digitalPinToInterrupt(DIT_PIN), ditPaddlePressedIRQ, LOW );
  attachInterrupt( digitalPinToInterrupt(DAH_PIN), dahPaddlePressedIRQ, LOW );
}

void checkButtonsAndKnobs()
{ 
  g_ButtonPressed = keyerIO.checkButtons();
  keyerIO.checkWPMPoti();
}

void enterCommandMode()
{
  g_IsInCommandMode = true;
  g_IsInProgrammingMode = false;
  g_CommandSequence = 0;
  digitalWrite( COMMAND_LED_PIN, true );
  keyerIO.bleepUp();
}

void leaveCommandMode()
{
  g_IsInCommandMode = false;
  g_IsInProgrammingMode = false;
  g_CommandSequence = 0;
  digitalWrite( COMMAND_LED_PIN, false );
  keyerIO.bleepDown();
}

void wakeOnButtonPress()
{
  if( g_ButtonPressed != 0 )
  {
    clock_prescale_set(clock_div_1);
    g_PowerCount = POWER_DOWN_TIME;
    g_SaveCount  = POWER_SAVE_TIME;
    digitalWrite( POWER_LED_PIN, true );
  }
}

void checkCommandButtonPressed()
{
  if( g_ButtonPressed == 1 )
  {
    if( g_IsInCommandMode )
    {
      leaveCommandMode();
    }
    else
    {
      enterCommandMode();
    }
    // wait for the user to release any button if pressed
    do
    { 
      delay(10);
    }
    while( keyerIO.checkButtons() != 0 );
    g_ButtonPressed = 0;
  }
}

void checkMemoryButtonPressed()
{
  if( !g_IsInCommandMode && g_ButtonPressed != 0 )
  {
    // MemButton pressed...        
    keyerMem.sendMemory( g_ButtonPressed - 2 );
    g_ButtonPressed = 0;
  }
}

void checkFunctionButtonPressed()
{
  if( g_IsInCommandMode && g_ButtonPressed != 0 )
  {
    switch( (g_ButtonPressed-1) )
    {
      case( 1 ):
      {
        // set 12 WPM
        keyerIO.setWPM( 12 );
        leaveCommandMode();
        break;
      }
      case( 2 ):
      {
        // set 15 WPM
        keyerIO.setWPM( 15 );
        leaveCommandMode();
        break;
      }
      case( 3 ):
      {
        // set 20 WPM
        keyerIO.setWPM( 20 );
        leaveCommandMode();
        break;
      }
      case( 4 ):
      {
        // set 25 WPM
        keyerIO.setWPM( 25 );
        leaveCommandMode();
        break;
      }
      case( 5 ):
      {
        // set Ratio to 1:3
        keyerIO.setRatio( 300 );
        leaveCommandMode();
        break;
      }
      case( 6 ):
      {
        // set Ratio to 1:3.25
        keyerIO.setRatio( 325, true );
        leaveCommandMode();
        break;
      }
      case( 7 ):
      {
        // set Ratio to 1:3.50
        keyerIO.setRatio( 350, true );
        leaveCommandMode();
        break;
      }
      case( 8 ):
      {
        // toggle SpeedWeight-Ratio On/Off
        keyerIO.toggleSpeedRatio();
        keyerIO.bleepOK();
        g_CommandSequence = 0;
        break;
      }
    }
    g_ButtonPressed = 0;
  }
}

void handleCommands()
{
  if( g_IsInCommandMode && g_ButtonPressed == 0 )
  {
    unsigned long int timeDelta = millis() - g_SymbolEndTime;
    
    if( timeDelta > ( keyerIO.getDitMillis() * 3 ) )
    {
      if( !g_IsInProgrammingMode )
      {
        switch( g_CommandSequence )
        {
          case 0:
          {
            break;
          }
          case 1: // "e" play wpm 
          {
            delay( keyerIO.getDitMillis() );
            keyerIO.sendNumberNoTX( keyerIO.getWPM() );
            break;
          }
          case 12: // "a" iambic mode A
          {
            keyerIO.setMode( KeyerIO::IAMBIC_A );
            keyerIO.bleepOK();
            break;
          }
          case 2111: // "b" iambic mode B
          {
            keyerIO.setMode( KeyerIO::IAMBIC_B );
            keyerIO.bleepOK();
            break;
          }
          case 211: // "d" ultimatic mode
          {
            keyerIO.setMode( KeyerIO::ULTIMATIC );
            keyerIO.bleepOK();
            break;
          }
          case 22: // "m" dot/dash-memory
          {
            keyerIO.toggleDotDashMemory();
            keyerIO.bleepOK();
            break;
          }
          case 1121: // "f" change frequency
          {
            keyerIO.bleepOK();
            bool notDone=true;
            while( notDone )
            {
              int const button = keyerIO.checkButtons();
              if( g_DitPressed )
              {
                keyerIO.setSideToneFrequency( keyerIO.getSideToneFrequency() + 5 );
                keyerIO.sendDit();
              }
              else
              if( g_DahPressed )
              {
                keyerIO.setSideToneFrequency( keyerIO.getSideToneFrequency() - 5 );
                keyerIO.sendDah();
              }
              if( button == 1 ) 
              {
                notDone = false;
                g_CommandSequence=0;
              }
            }
            delay( keyerIO.getDitMillis() );
            keyerIO.sendNumber( keyerIO.getSideToneFrequency() );
            break;
          }
          case 11: // "i" tx on/off
          {
            keyerIO.setTXState( !keyerIO.getTXState() );
            keyerIO.bleepOK();
            break;
          }
          case 222: // "o" turn side-tone on/off
          {
            keyerIO.toggleSideTone();
            keyerIO.bleepOK();
            break;
          }
          case 21: // "n" reverse paddles
          {
            keyerIO.toggleReversePaddle();
            keyerIO.bleepOK();
            break;
          }
          case 122: // "w" change speed
          {
            keyerIO.bleepOK();
            bool notDone=true;
            while( notDone )
            {
              int const button = keyerIO.checkButtons();
              if( g_DitPressed )
              {
                keyerIO.setWPM( keyerIO.getWPM() + 1 );
                keyerIO.sendDit();
              }
              else
              if( g_DahPressed )
              {
                keyerIO.setWPM( keyerIO.getWPM() - 1 );
                keyerIO.sendDah();
              }
              if( button == 1 ) 
              {
                notDone = false;
                g_CommandSequence=0;
              }
            }
            delay( keyerIO.getDitMillis() );
            keyerIO.sendNumberNoTX( keyerIO.getWPM() );
            break;
          }
          case 1221: // "p" start programming
          {
            g_IsInProgrammingMode = true;
            keyerIO.bleepOK();
            break;
          }
          case 2112: // "x" leave Command-Mode
          {
            g_IsInProgrammingMode = false;
            g_IsInCommandMode = false;
            digitalWrite( COMMAND_LED_PIN, false );
            keyerIO.bleepDown();
            break;
          }
          case 2:
          {
            while(true)
            {
              int const button = keyerIO.checkButtons();
              if( button == 1 ) break; // leave this

              bool const state = (!digitalRead( DIT_PIN ))| 
                                 (!digitalRead( DAH_PIN ));

              if( state == false )
              {
                digitalWrite( OUT_PIN, false );
                noTone( BEEPER_PIN );
              }
              else
              {
                if( keyerIO.isSideToneOn() )
                {
                  tone( BEEPER_PIN, keyerIO.getSideToneFrequency() );
                }
                digitalWrite( OUT_PIN, true );
              }
              
            }
            break;
          }
          default:
          {
            keyerIO.bleepError();
            g_CommandSequence = 0;
            break;
          }
        }
      }
      else
      {
        switch( g_CommandSequence )
        {
          case 0:
          {
            break;
          }
          case 12222: // Memory 1 selected
          {
            keyerIO.bleepOK();
            keyerMem.updateLoop(0);
            keyerIO.bleepOK();
            break;
          }
          case 11222: // Memory 2 selected
          {
            keyerIO.bleepOK();
            keyerMem.updateLoop(1);
            keyerIO.bleepOK();
            break;
          }
          case 11122: // Memory 3 selected
          {
            keyerIO.bleepOK();
            keyerMem.updateLoop(2);
            keyerIO.bleepOK();
            break;
          }
          case 11112: // Memory 4 selected
          {
            keyerIO.bleepOK();
            keyerMem.updateLoop(3);
            keyerIO.bleepOK();
            break;
          }
          case 11111: // Memory 5 selected
          {
            keyerIO.bleepOK();
            keyerMem.updateLoop(4);
            keyerIO.bleepOK();
            break;
          }
          case 21111: // Memory 6 selected
          {
            keyerIO.bleepOK();
            keyerMem.updateLoop(5);
            keyerIO.bleepOK();
            break;
          }
          case 22111: // Memory 7 selected
          {
            keyerIO.bleepOK();
            keyerMem.updateLoop(6);
            keyerIO.bleepOK();
            break;
          }
          case 22211: // Memory 8 selected
          {
            keyerIO.bleepOK();
            keyerMem.updateLoop(7);
            keyerIO.bleepOK();
            break;
          }
          default:
          {
            g_IsInProgrammingMode = false;
            keyerIO.bleepError();
            break;
          }
        }
      }
      
      // no matter what, clear commandSequence
      g_CommandSequence = 0;
    }
    g_ButtonPressed = 0;
  }
}

void loop()
{
  checkButtonsAndKnobs();
  
  wakeOnButtonPress();
  checkCommandButtonPressed();
  checkMemoryButtonPressed();
  checkFunctionButtonPressed();

  handleCommands();

  
  if( g_DitPressed || g_DahPressed )
  {  
    if( g_DitPressed && !g_DahPressed )
    {
      keyerIO.sendDit();
    }
    else
    if( g_DahPressed && !g_DitPressed )
    {
      keyerIO.sendDah();
    }
    else
    if( g_DitPressed && g_DahPressed )
    {
      keyerIO.sendAlternating();
    }
  }
  else
  {
    if( !g_IsInCommandMode )
    {
      if( g_SaveCount <= 0 )
      { 
        static bool ledCnt = false;
        ledCnt = !ledCnt;
        clock_prescale_set(clock_div_128);
        digitalWrite( POWER_LED_PIN, ledCnt );
        
        if( g_PowerCount <= 0 )
        {
          clock_prescale_set(clock_div_1);
          sleep_enable();
          digitalWrite( POWER_LED_PIN, false );
          set_sleep_mode(SLEEP_MODE_PWR_DOWN);
          sleep_cpu();
        }
        else
        {
          g_PowerCount-=128;
          delayMicroseconds( 700 );
        }
      }
      else
      {
        g_PowerCount = POWER_DOWN_TIME;
        g_SaveCount-=1;
        delayMicroseconds( 700 );
      }
    }
    
    keyerIO.eepromStoreWpm();
    keyerIO.eepromStoreSideTone();
    keyerIO.eepromStoreReversePaddle();
    keyerIO.eepromStoreSpeedRatio();
    keyerIO.eepromStoreFrequency();
  }
  
}
