#include "keyerio.h"
#include "morse.h"

#include <arduino.h>

extern volatile int g_PowerCount;
extern volatile int g_SaveCount;

extern volatile bool g_DitPressed;
extern volatile bool g_DahPressed;

extern volatile bool g_IsInCommandMode;
extern volatile unsigned int g_CommandSequence;
extern volatile unsigned long int g_SymbolEndTime;

extern KeyerIO keyerIO;
extern Morse morse;


void ditPaddlePressedIRQ()
{
  cli();
  sleep_disable();
  clock_prescale_set(clock_div_1);
  g_PowerCount = POWER_DOWN_TIME;
  g_SaveCount = POWER_SAVE_TIME;
  
  if( !keyerIO.isReversePaddle() )
  {
    g_DitPressed = true;
  }
  else
  {
    g_DahPressed = true;
  }
  
  detachInterrupt( digitalPinToInterrupt(DIT_PIN) );
  digitalWrite( POWER_LED_PIN, true );
  sei();
}

void dahPaddlePressedIRQ()
{
  cli();
  sleep_disable();
  clock_prescale_set(clock_div_1);
  g_PowerCount = POWER_DOWN_TIME;
  g_SaveCount = POWER_SAVE_TIME;
  
  if( !keyerIO.isReversePaddle() )
  {
    g_DahPressed = true;
  }
  else
  {
    g_DitPressed = true;
  }
  
  detachInterrupt( digitalPinToInterrupt(DAH_PIN) );
  digitalWrite( POWER_LED_PIN, true );
  sei();
}


void KeyerIO::setRatio( unsigned int ratio, bool correctForWPM )
{
  m_Ratio = ratio;

  m_DitMillis = 60000 / ( m_CurrentWPM * 50 );
  m_DahMillis = (m_DitMillis * m_Ratio) / 100;  

  if( correctForWPM )
  {
    long int const s = 20*100 + 10*m_Ratio;
    m_DitMillis = m_DitMillis * 5000 / s;
    m_DahMillis = m_DahMillis * 5000 / s;
  }
}

void KeyerIO::setWPM( unsigned int speedWPM )
{
  if( speedWPM > MAX_WPM )
  {
    speedWPM = MAX_WPM;
  }
  else
  if( speedWPM < MIN_WPM )
  {
    speedWPM = MIN_WPM;
  }
  
  m_CurrentWPM = speedWPM;

  if( m_UseSpeedRatio )
  {
    float blend = ( (float)m_CurrentWPM - (float)MIN_WPM ) /
                  ( (float)25.f - (float)MIN_WPM );
    blend = blend > 1.f ? 1.f:blend;
    m_Ratio = 300 * (1.f-blend) + 350 * blend;
  }
  
  m_DitMillis = 60000 / ( m_CurrentWPM * 50 );
  m_DahMillis = (m_DitMillis * m_Ratio) / 100;
}

unsigned int KeyerIO::getDitMillis()
{
  return( m_DitMillis );
}

unsigned int KeyerIO::getDahMillis()
{
  return( m_DahMillis );
}

unsigned char KeyerIO::getWPM()
{
  return( m_CurrentWPM );
}

void KeyerIO::init( bool reset )
{
  // init WPM-Poti, so its current value does not interfere
  // with the following settings
  checkWPMPoti( true );

  if( reset )
  {
    // first start after flashing firmware or reset, so set defaults
    EEPROM.write( EEPROM_WPM_ADDRESS, DEFAULT_WPM );
    EEPROM.write( EEPROM_SIDETONE_FREQUENCY_ADDRESS, (SIDETONE_DEFAULT_FREQUENCY-225)/5 );
    EEPROM.write( EEPROM_SIDETONE_ADDRESS, SIDETONE_DEFAULT_ON? 1:0 );
    EEPROM.write( EEPROM_REVERSE_PADDLE_ADDRESS, REVERSE_PADDLE_DEFAULT_ON? 1:0);
    EEPROM.write( EEPROM_SPEED_RATIO_ADDRESS, SPEED_RATIO_DEFAULT_ON? 1:0 );
    EEPROM.write( EEPROM_MODE_ADDRESS, 0 );
    EEPROM.write( EEPROM_DOT_DASH_MEMORY_ADDRESS, 1 );
  }

  setWPM( EEPROM.read( EEPROM_WPM_ADDRESS ) );
  setRatio( 300 );
  setSideToneFrequency( EEPROM.read( EEPROM_SIDETONE_FREQUENCY_ADDRESS )*5+225 );
  toggleSideTone( EEPROM.read( EEPROM_SIDETONE_ADDRESS ) == 1 );
  toggleReversePaddle( EEPROM.read( EEPROM_REVERSE_PADDLE_ADDRESS ) == 1 );
  toggleSpeedRatio( EEPROM.read( EEPROM_SPEED_RATIO_ADDRESS ) == 1 );
  toggleDotDashMemory( EEPROM.read( EEPROM_DOT_DASH_MEMORY_ADDRESS ) == 1 );
  setMode( EEPROM.read( EEPROM_MODE_ADDRESS ) == 0 ? IAMBIC_A:
           EEPROM.read( EEPROM_MODE_ADDRESS ) == 1 ? IAMBIC_B:
           EEPROM.read( EEPROM_MODE_ADDRESS ) == 2 ? CMOS:
                                                     ULTIMATIC );
           
}

void KeyerIO::setSideToneFrequency( unsigned int frequency )
{
  frequency = frequency <  225 ?  225:frequency;
  frequency = frequency > 1500 ? 1500:frequency;
  m_SideToneFrequency = frequency;
}

unsigned int KeyerIO::getSideToneFrequency()
{
  return( m_SideToneFrequency );
}

void KeyerIO::sendSymbol( bool isDit )
{  
  if( m_Mode == IAMBIC_B )
  {
    //attachInterrupt( digitalPinToInterrupt(DAH_PIN), dahPaddlePressedIRQ, LOW );
    //attachInterrupt( digitalPinToInterrupt(DIT_PIN), ditPaddlePressedIRQ, LOW );
  }
  
  unsigned int const symbolTime = isDit? m_DitMillis:m_DahMillis;
  unsigned int const pauseTime = m_DitMillis;

  // turn beeper on if sidetone is turned on or if keyer is
  // in command-mode

  if( g_IsInCommandMode || 
      m_SideToneOn    ||
      m_DoNotTransmit )
  {
    tone( BEEPER_PIN, m_SideToneFrequency );
  }

  if( !g_IsInCommandMode )
  {
    // if not in command-mode turn tx-out-pin high
    if( !m_DoNotTransmit )
    {
      digitalWrite( OUT_PIN, true );
    }
  }
  else
  {
    // if in command-mode update command-sequence

    g_CommandSequence = g_CommandSequence * 10;

    if( isDit )
    {
      g_CommandSequence += 1;
    }
    else
    {
      g_CommandSequence += 2;
    }
  }

  delay( symbolTime );
  g_SymbolEndTime = millis();

  if( !g_IsInCommandMode )
  {
    // if not in command-mode turn tx-out-pin low
    digitalWrite( OUT_PIN, false);
  }

  // if beeper was turned on, turn it off again
  
  noTone( BEEPER_PIN );
  
  // after the pause-time turn the IRQ for the current symbol on again
  
  delay( pauseTime );

  // if requested by the user, that there shall be no dot-/dash-memory,
  // clear stored dots and dashes in there...
  
  if(!m_DotDashMemory)
  {
    g_DitPressed=false;
    g_DahPressed=false;
  }
      
  attachInterrupt( digitalPinToInterrupt(DIT_PIN), ditPaddlePressedIRQ, LOW );
  attachInterrupt( digitalPinToInterrupt(DAH_PIN), dahPaddlePressedIRQ, LOW );
}

void KeyerIO::sendDit()
{
  g_DitPressed=false;
  sendSymbol( true );
  m_LastSymbolWasDit = true;
}

void KeyerIO::sendDah()
{
  g_DahPressed=false;
  sendSymbol( false );
  m_LastSymbolWasDit = false;
}

void KeyerIO::sendAlternating()
{
  // if both paddles pressed simultanously ("iambic")
  // we need to reset both for a propper refresh
  g_DitPressed=false;
  g_DahPressed=false;
  
  if( m_LastSymbolWasDit )
  {
    sendDah();
    if( m_Mode == IAMBIC_B && !g_DitPressed && !g_DahPressed )
    {
      sendDit();
    }
    else
    if( m_Mode == ULTIMATIC )
    {
      while( g_DahPressed )
      {
        g_DitPressed=false;
        g_DahPressed=false;
        sendDah();
      }
    }
  }
  else
  {
    sendDit();
    if( m_Mode == IAMBIC_B && !g_DitPressed && !g_DahPressed )
    {
      sendDah();
    }
    else
    if( m_Mode == ULTIMATIC )
    {
      while( g_DitPressed )
      {
        g_DitPressed=false;
        g_DahPressed=false;
        sendDit();
      }
    }
  }
}

void KeyerIO::sendChar( unsigned char c )
{
  unsigned char bits = morse.toMorse( c ); //pgm_read_byte(&ASCII2Morse[c&0xff]);

  if( bits != 0x00 )
  {
    if( bits > 1 )
    do
    {
      if( g_DitPressed || g_DahPressed ) return;
  
      if( bits & 1 )
      {
        keyerIO.sendDit();
      }
      else
      {
        keyerIO.sendDah();
      }
  
      bits >>= 1;
    }
    while( bits > 1 );
    
    delay( m_DahMillis );
  }
}

void KeyerIO::sendString( unsigned char const * const string )
{
  int i=0;
  do 
  {
    if( g_DitPressed || g_DahPressed ) return;

    char c = string[i];

    if( c == ' ' )
    {
      delay( m_DahMillis );
      if( g_DitPressed || g_DahPressed ) return;
      delay( m_DahMillis );
      if( g_DitPressed || g_DahPressed ) return;
      delay( m_DitMillis );
    }
    else
    {
      sendChar( c );
    }

    i++;
  }
  while( string[i] != 0 );
}

void KeyerIO::sendStringNoTX( unsigned char const * const string )
{
  bool oldState = m_DoNotTransmit;
  m_DoNotTransmit = true;
  sendString( string );
  m_DoNotTransmit = oldState;
}

void KeyerIO::sendNumber( int number )
{
  bool previousNumberSent = false;
  
  for( int div = 10000; div > 0; div /= 10 )
  {
    if( g_DitPressed || g_DahPressed ) break;
    
    char const digit = ( number / div ) % 10;

    if( digit != 0 || div == 1 || previousNumberSent )
    {
      sendChar( '0'+digit );
      previousNumberSent = true;
    }
  }
}

void KeyerIO::sendNumberNoTX( int number )
{
  bool oldState = m_DoNotTransmit;
  m_DoNotTransmit = true;
  sendNumber( number );
  m_DoNotTransmit = oldState;
}

void KeyerIO::bleepUp()
{
  tone( BEEPER_PIN, TONE_LOW );
  delay( 100 );
  tone( BEEPER_PIN, TONE_MID );
  delay( 100 );
  tone( BEEPER_PIN, TONE_HIG );
  delay( 100 );
  noTone( BEEPER_PIN );
}

void KeyerIO::bleepDown()
{
  tone( BEEPER_PIN, TONE_HIG );
  delay( 100 );
  tone( BEEPER_PIN, TONE_MID );
  delay( 100 );
  tone( BEEPER_PIN, TONE_LOW );
  delay( 100 );
  noTone( BEEPER_PIN );
}

void KeyerIO::bleepOK()
{
  delay( 100 );
  tone( BEEPER_PIN, TONE_HIG );
  delay( 100 );
  noTone( BEEPER_PIN );
}

void KeyerIO::bleepError()
{
  delay( 20 );
  tone( BEEPER_PIN, TONE_LOW );
  delay( 20 );
  tone( BEEPER_PIN, TONE_LOW );
  delay( 20 );
  tone( BEEPER_PIN, TONE_LOW );
  delay( 20 );
  tone( BEEPER_PIN, TONE_LOW );
  delay( 20 );
  tone( BEEPER_PIN, TONE_LOW );
  delay( 20 );
  tone( BEEPER_PIN, TONE_LOW );
  delay( 20 );
  tone( BEEPER_PIN, TONE_LOW );
  delay( 20 );
  tone( BEEPER_PIN, TONE_LOW );
  delay( 20 );
  tone( BEEPER_PIN, TONE_LOW );
  delay( 20 );
  tone( BEEPER_PIN, TONE_LOW );
  delay( 100 );
  noTone( BEEPER_PIN );
}

void KeyerIO::toggleSideTone()
{
  m_SideToneOn = !m_SideToneOn;
}

void KeyerIO::toggleSideTone( bool state )
{
  m_SideToneOn = state;
}

void KeyerIO::toggleReversePaddle()
{
  m_ReversePaddle = !m_ReversePaddle;
}

void KeyerIO::toggleReversePaddle( bool state )
{
  m_ReversePaddle = state;
}

void KeyerIO::toggleSpeedRatio()
{
  m_UseSpeedRatio = !m_UseSpeedRatio;
}

void KeyerIO::toggleSpeedRatio( bool state )
{
  m_UseSpeedRatio = state;  
}

void KeyerIO::toggleDotDashMemory()
{
  m_DotDashMemory = !m_DotDashMemory;
}

void KeyerIO::toggleDotDashMemory( bool state )
{
  m_DotDashMemory = state;
}

bool KeyerIO::isReversePaddle()
{
  return( m_ReversePaddle );
}

bool KeyerIO::isSideToneOn()
{
  return( m_SideToneOn );
}

bool KeyerIO::isSpeedRatioOn()
{
  return( m_UseSpeedRatio );
}

bool KeyerIO::isDotDashMemoryOn()
{
  return( m_DotDashMemory );
}

int KeyerIO::checkButtons()
{  
  float value = (float)analogRead( BUTTONS_PIN );

  int button = 20;

  if( value != 1023.f )
  {
    button = (int)floor( - ( 10.f * value ) / ( value - 1023.f ) + 0.5f );
    button = button < 0  ?  0:button;
    button = button > 20 ? 20:button;
  }
  
  if( button == 20 ) 
  {
    return( 0 );
  }
  
  return( button + 1 );
}

void KeyerIO::checkWPMPoti( bool init )
{
  // When initializing, then do everything as usual but do *not*
  // really alter the speed. This is just done to set up the 
  // static variable indicating a Potentiometer-Movement.
  
  static unsigned int lastPotiWpm = DEFAULT_WPM;

  unsigned int potiValue = analogRead( WPM_POTI_PIN );
  unsigned int newWpm = ( POTI_MAX_WPM*potiValue + POTI_MIN_WPM*(1023-potiValue) ) / 1023;
  
  if( init )
  {
    lastPotiWpm = newWpm;
    return;
  }
  
  if( newWpm != lastPotiWpm )
  {
    lastPotiWpm = newWpm;
    setWPM( newWpm );
    clock_prescale_set(clock_div_1);
    g_PowerCount = POWER_DOWN_TIME;
    g_SaveCount  = POWER_SAVE_TIME;
    digitalWrite( POWER_LED_PIN, true );
  }
}


bool KeyerIO::getTXState()
{
  return( !m_DoNotTransmit );
}

void KeyerIO::setTXState( bool state )
{
  m_DoNotTransmit = !state;
}

void KeyerIO::eepromStoreWpm()
{
  // every 3 seconds check if the wpm setting 
  // differs from the value stored in the eeprom
  // and if so update it
  
  static unsigned long int lastUpdateTime = millis();
  unsigned long int timeDelta = millis()-lastUpdateTime;

  if( timeDelta > 1000 )
  {
    lastUpdateTime = millis();

    unsigned char eepromWpm = EEPROM.read( EEPROM_WPM_ADDRESS );
    if( eepromWpm != m_CurrentWPM )
    {
      EEPROM.write( EEPROM_WPM_ADDRESS, m_CurrentWPM );
    }
  }
}

void KeyerIO::eepromStoreSideTone()
{
  // every 3 seconds check if the sidetone setting 
  // differs from the value stored in the eeprom
  // and if so update it

  static unsigned long int lastUpdateTime = millis();
  unsigned long int timeDelta = millis()-lastUpdateTime;

  if( timeDelta > 1000 )
  {
    lastUpdateTime = millis();

    bool eepromSideTone = EEPROM.read( EEPROM_SIDETONE_ADDRESS )==1;
    if( eepromSideTone != m_SideToneOn )
    {
      EEPROM.write( EEPROM_SIDETONE_ADDRESS, m_SideToneOn? 1:0 );
    }
  }
}

void KeyerIO::eepromStoreFrequency()
{
  // every 3 seconds check if the sidetone-frequency setting 
  // differs from the value stored in the eeprom
  // and if so update it
  
  static unsigned long int lastUpdateTime = millis();
  unsigned long int timeDelta = millis()-lastUpdateTime;

  if( timeDelta > 1000 )
  {
    lastUpdateTime = millis();

    unsigned char eepromFreq = EEPROM.read( EEPROM_SIDETONE_FREQUENCY_ADDRESS );
    if( ((unsigned int)eepromFreq*5+225) != m_SideToneFrequency )
    {
      EEPROM.write( EEPROM_SIDETONE_FREQUENCY_ADDRESS, (m_SideToneFrequency-225)/5 );
    }
  }
}

void KeyerIO::eepromStoreReversePaddle()
{
  // every 3 seconds check if the reverse-paddle setting 
  // differs from the value stored in the eeprom
  // and if so update it
  
  static unsigned long int lastUpdateTime = millis();
  unsigned long int timeDelta = millis()-lastUpdateTime;

  if( timeDelta > 1000 )
  {
    lastUpdateTime = millis();

    unsigned char reversePaddle = EEPROM.read( EEPROM_REVERSE_PADDLE_ADDRESS )==1;
    if( reversePaddle != m_ReversePaddle )
    {
      EEPROM.write( EEPROM_REVERSE_PADDLE_ADDRESS, m_ReversePaddle? 1:0 );
    }
  }
}

void KeyerIO::eepromStoreSpeedRatio()
{
  // every 3 seconds check if the speed-ratio setting 
  // differs from the value stored in the eeprom
  // and if so update it
  
  static unsigned long int lastUpdateTime = millis();
  unsigned long int timeDelta = millis()-lastUpdateTime;

  if( timeDelta > 1000 )
  {
    lastUpdateTime = millis();

    unsigned char speedRatio = EEPROM.read( EEPROM_SPEED_RATIO_ADDRESS )==1;
    if( speedRatio != m_UseSpeedRatio )
    {
      EEPROM.write( EEPROM_SPEED_RATIO_ADDRESS, m_UseSpeedRatio? 1:0 );
    }
  }
}

void KeyerIO::setMode( enum MODE mode )
{
  m_Mode = mode;
}


