
#include "memory.h"
#include "morse.h"
#include "keyerio.h"

extern volatile bool g_DitPressed;
extern volatile bool g_DahPressed;
extern volatile unsigned long int g_SymbolEndTime;
extern volatile unsigned int g_CommandSequence;

extern Morse morse;
extern KeyerIO keyerIO;

void Memory::init( bool reset )
{ 
  if( reset )
  {
    // OK, it seems like the keyer was never used before or shall
    // be reset, so init the memory-keys with some texts
    setString( 0, F(DEFAULT_MSG_0) );
    setString( 1, F(DEFAULT_MSG_1) );
    setString( 2, F(DEFAULT_MSG_2) );
    setString( 3, F(DEFAULT_MSG_3) );
    setString( 4, F(DEFAULT_MSG_4) );
    setString( 5, F(DEFAULT_MSG_5) );
    setString( 6, F(DEFAULT_MSG_6) );
    setString( 7, F(DEFAULT_MSG_7) );
  }
  
  m_MemIndex = 0;
  m_Position = 0;
}

void Memory::setMemory( int index )
{
  m_MemIndex = index;
  m_Position = 0;
  m_Bits = 0;
}

bool Memory::getMemBit()
{
  int const bytePos = m_Position >> 3;
  int const bitPos  = 7 - ( m_Position & 7 );
  int const addr = EEPROM_MEM_OFFSET + MEM_MAX_SIZE*m_MemIndex + bytePos;
  return( ( EEPROM.read( addr ) >> bitPos ) & 1 );
}

void Memory::setMemBit( bool bit )
{
  int const bytePos = m_Position >> 3;
  int const bitPos  = 7 - ( m_Position & 7 );

  if( bit )
  {
    // set bit in temp bit-buffer
    m_Bits |= 1 << bitPos;
  }
  else
  {
    // clear bit in temp bit-buffer
    m_Bits &= ~(1 << bitPos);
  }

  if( bitPos == 0 ) // last bit of 8?
  {
    // yes... write full byte into eeprom
    int const addr = EEPROM_MEM_OFFSET + MEM_MAX_SIZE*m_MemIndex + bytePos;
    EEPROM.write( addr, m_Bits );
  }
}

void Memory::flushMemBits()
{
  int const bytePos = m_Position >> 3;
  int const bitPos  = 7 - ( m_Position & 7 );
  int const addr = EEPROM_MEM_OFFSET + MEM_MAX_SIZE*m_MemIndex + bytePos;
  EEPROM.write( addr, m_Bits );  
}

Memory::SYMBOLS Memory::getSymbol()
{
  enum SYMBOLS symbol = END;
  
  bool bit = getMemBit();
  m_Position++;
  
  if( bit )
  {
    symbol = DIT;
  }
  else
  {
    bit = getMemBit();
    m_Position++;

    if( bit )
    {
      symbol = DAH;
    }
    else
    {
      unsigned char nextBits = 0;
      nextBits |= getMemBit()? (1 << 0) : 0;
      m_Position++;
      nextBits |= getMemBit()? (1 << 1) : 0;
      m_Position++;
      nextBits |= getMemBit()? (1 << 2) : 0;
      m_Position++;
      nextBits |= getMemBit()? (1 << 3) : 0;

      if( (nextBits & 0x0f) == 0x00 )
      {
        symbol = END;
      }
      else
      if( (nextBits & 0x03) == 0x00 )
      {
        m_Position -= 1;
        symbol = WPAUSE;
      }
      else
      {
        m_Position -= 3;
        symbol = CPAUSE;
      }
    }
  }
  
  return( symbol );
}

bool Memory::addSymbol( int symbol )
{
  // enough bits left to store something?
  if( m_Position > ( MEM_MAX_BITS - 8 ) )
  {
    return( false );
  }
  
  switch( symbol )
  {
    case 1: // dit
    {
        setMemBit( true );
        m_Position++;
        break;
    }
    case 2: // dah
    {
        setMemBit( false );
        m_Position++;
        setMemBit( true );
        m_Position++;
        break;
    }
    case 3: // pause
    {
        setMemBit( false );
        m_Position++;
        setMemBit( false );
        m_Position++;
        break;
    }
  }
  
  return( true );
}

void Memory::addChar( char c )
{
  unsigned char bits = morse.toMorse( c );//pgm_read_word(&ASCII2Morse[c&0xff]);

  if( bits != 0x00 )
  {
    if( bits > 1 )
    do
    {
      if( bits & 1 )
      {
        addSymbol( 1 );
      }
      else
      {
        addSymbol( 2 );
      }
  
      bits >>= 1;
    }
    while( bits > 1 );
    
    addSymbol( 3 );
  }
}

void Memory::setString( int memIndex, char const * const string )
{
  setMemory( memIndex );
  
  int i=0;
  do 
  {
    char c = string[i];

    if( c == ' ' )
    {
      addSymbol( 3 );
    }
    else
    if( c == '/' )
    {
      addSymbol( 2 );
      addSymbol( 1 );
      addSymbol( 1 );
      addSymbol( 1 );
      addSymbol( 2 );
      addSymbol( 3 );
    }
    else
    {
      addChar( c );
    }

    i++;
  }
  while( string[i] != 0 );

  // terminate input
  addSymbol( 3 );
  addSymbol( 3 );
  flushMemBits();
}

void Memory::setString( int memIndex, __FlashStringHelper const * const data )
{
  const char PROGMEM *p = (const char PROGMEM *)data;
  setMemory( memIndex );
  
  int i=0;
  do 
  {
    char c = pgm_read_byte( &p[i] );

    if( c == ' ' )
    {
      addSymbol( 3 );
    }
    else
    if( c == '/' )
    {
      addSymbol( 2 );
      addSymbol( 1 );
      addSymbol( 1 );
      addSymbol( 1 );
      addSymbol( 2 );
      addSymbol( 3 );
    }
    else
    {
      addChar( c );
    }

    i++;
  }
  while( pgm_read_byte( &p[i] ) != 0 );

  // terminate input
  addSymbol( 3 );
  addSymbol( 3 );
  flushMemBits();
}

void Memory::sendMemory( int memIndex )
{
  
  bool done = false;
  
  setMemory( memIndex );
  
  do
  {
    keyerIO.checkWPMPoti();
  
    enum SYMBOLS symbol = getSymbol();
  
    switch( symbol )
    {
      case( Memory::DIT ):
      {
        keyerIO.sendDit();
        break;
      }
      case( Memory::DAH ):
      {
        keyerIO.sendDah();
        break;
      }
      case( Memory::CPAUSE ):
      {
        delay( 2 * keyerIO.getDitMillis() );
        break;
      }
      case( Memory::WPAUSE ):
      {
        delay( 6 * keyerIO.getDitMillis() );
        break;
      }
      case( Memory::END ):
      {
        done = true;
        break;
      }
    }
    
    if( g_DitPressed || g_DahPressed )
    {
      done = true;
    }
  }
  while( !done );
}

void Memory::updateLoop( int memIndex )
{
  bool notDone = true;
  
  setMemory( memIndex );

  bool newCharSent = false;
  bool newWordSent = false;
  bool memoryFull  = false;
  
  while( notDone )
  {
    unsigned long int timeDelta = millis() - g_SymbolEndTime;
    
    if( timeDelta > (2*keyerIO.getDitMillis()) && newCharSent == false )
    {
      memoryFull  = !addSymbol(3);
      newCharSent = true;
    }

    if( timeDelta > (5*keyerIO.getDitMillis()) && newWordSent == false )
    {
      memoryFull  = !addSymbol(3);
      newWordSent = true;
    }
    
    if( g_DitPressed || g_DahPressed )
    {  
      if( g_DitPressed && !g_DahPressed )
      {
        keyerIO.sendDit();
        memoryFull  = !addSymbol(1);
        newCharSent = false;
        newWordSent = false;
      }
      else
      if( g_DahPressed && !g_DitPressed )
      {
        keyerIO.sendDah();
        memoryFull  = !addSymbol(2);
        newCharSent = false;
        newWordSent = false;
      }
      else
      if( g_DitPressed && g_DahPressed )
      {
        keyerIO.sendAlternating();
      }
    }
    
    if( keyerIO.checkButtons() == 1 || memoryFull )
    {
      notDone = false;
      addSymbol(3);
      addSymbol(3);
      addSymbol(3);
      flushMemBits();
    }
  }
  g_CommandSequence = 0;
}

